@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2012, Canonical Ltd
+# Copyright (c) 2015 Will Bond, Mjumbe Wawatu Ukweli, 2012 Canonical Ltd
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -16,20 +16,23 @@
 
 """The compiler for pybars."""
 
+import re
+import sys
+from types import ModuleType
+import linecache
+
+import pybars
+import pybars._templates
+from pymeta.grammar import OMeta
 
 __all__ = [
     'Compiler',
     'strlist',
+    'Scope'
     ]
 
 __metaclass__ = type
 
-from functools import partial
-import re
-
-import pybars
-from pymeta.grammar import OMeta
-from pybars.frompymeta import moduleFromSource
 
 # This allows the code to run on Python 2 and 3 by
 # creating a consistent reference for the appropriate
@@ -40,10 +43,10 @@ except NameError:
     # Python 3 support
     str_class = str
 
-import collections
 
-# preserve reference to the builtin compile.
-_compile = compile
+# Flag for testing
+debug = False
+
 
 # Note that unless we presume handlebars is only generating valid html, we have
 # to accept anything - so a broken template won't be all that visible - it will
@@ -53,8 +56,11 @@ _compile = compile
 handlebars_grammar = r"""
 
 template ::= (<text> | <templatecommand>)*:body => ['template'] + body
-text ::= (~(<start>) <anything>)+:text => ('literal', u''.join(text))
-other ::= <anything>:char => ('literal', char)
+text ::= <newline_text> | <whitespace_text> | <other_text>
+newline_text ::= (~(<start>) ('\r'?'\n'):char) => ('newline', u'' + char)
+whitespace_text ::= (~(<start>) (' '|'\t'))+:text => ('whitespace', u''.join(text))
+other_text ::= (~(<start>) <anything>)+:text => ('literal', u''.join(text))
+other ::= <anything>:char => ('literal', u'' + char)
 templatecommand ::= <blockrule>
     | <comment>
     | <escapedexpression>
@@ -64,7 +70,8 @@ start ::= '{' '{'
 finish ::= '}' '}'
 comment ::= <start> '!' (~(<finish>) <anything>)* <finish> => ('comment', )
 space ::= ' '|'\t'|'\r'|'\n'
-arguments ::= (<space>+ (<kwliteral>|<literal>|<path>))*:arguments => arguments
+arguments ::= (<space>+ (<kwliteral>|<literal>|<path>|<subexpression>))*:arguments => arguments
+subexpression ::= '(' <spaces> <path>:p (<space>+ (<kwliteral>|<literal>|<path>|<subexpression>))*:arguments <spaces> ')' => ('subexpr', p, arguments)
 expression_inner ::= <spaces> <path>:p <arguments>:arguments <spaces> <finish> => (p, arguments)
 expression ::= <start> '{' <expression_inner>:e '}' => ('expand', ) + e
     | <start> '&' <expression_inner>:e => ('expand', ) + e
@@ -72,18 +79,20 @@ escapedexpression ::= <start> <expression_inner>:e => ('escapedexpand', ) + e
 block_inner ::= <spaces> <symbol>:s <arguments>:args <spaces> <finish>
     => (u''.join(s), args)
 partial_inner ::= <spaces> <partialname>:s <arguments>:args <spaces> <finish>
-    => (u''.join(s), args)
+    => (s, args)
 alt_inner ::= <spaces> ('^' | 'e' 'l' 's' 'e') <spaces> <finish>
 partial ::= <start> '>' <partial_inner>:i => ('partial',) + i
 path ::= ~('/') <pathseg>+:segments => ('path', segments)
-kwliteral ::= <safesymbol>:s '=' (<literal>|<path>):v => ('kwparam', s, v)
-literal ::= (<string>|<integer>|<boolean>):thing => ('literalparam', thing)
+kwliteral ::= <safesymbol>:s '=' (<literal>|<path>|<subexpression>):v => ('kwparam', s, v)
+literal ::= (<string>|<integer>|<boolean>|<null>|<undefined>):thing => ('literalparam', thing)
 string ::= '"' <notdquote>*:ls '"' => u'"' + u''.join(ls) + u'"'
     | "'" <notsquote>*:ls "'" => u"'" + u''.join(ls) + u"'"
 integer ::= '-'?:sign <digit>+:ds => int((sign if sign else '') + ''.join(ds))
 boolean ::= <false>|<true>
 false ::= 'f' 'a' 'l' 's' 'e' => False
 true ::= 't' 'r' 'u' 'e' => True
+null ::= ('n' 'u' 'l' 'l') => None
+undefined ::= ('u' 'n' 'd' 'e' 'f' 'i' 'n' 'e' 'd') => None
 notdquote ::= <escapedquote>
     | '\n' => '\\n'
     | '\r' => '\\r'
@@ -99,7 +108,8 @@ escapedquote ::= '\\' '"' => '\\"'
 notclosebracket ::= (~(']') <anything>)
 safesymbol ::=  ~<alt_inner> '['? (<letter>|'_'):start (<letterOrDigit>|'_')+:symbol ']'? => start + u''.join(symbol)
 symbol ::=  ~<alt_inner> '['? (<letterOrDigit>|'-'|'@')+:symbol ']'? => u''.join(symbol)
-partialname ::= ~<alt_inner> ('['|'"')? (~(<space>|<finish>|']'|'"' ) <anything>)+:symbol (']'|'"')? => u''.join(symbol)
+partialname ::= <subexpression>:s => s
+    | ~<alt_inner> ('['|'"')? (~(<space>|<finish>|']'|'"' ) <anything>)+:symbol (']'|'"')? => ('literalparam', '"' + u''.join(symbol) + '"')
 pathseg ::= '[' <notclosebracket>+:symbol ']' => u''.join(symbol)
     | ('@' '.' '.' '/') => u'@@_parent'
     | <symbol>
@@ -111,7 +121,7 @@ symbolfinish :expected ::= <start> '/' <symbol>:found ?(found == expected) <fini
 blockrule ::= <start> '#' <block_inner>:i
       <template>:t <alttemplate>:alt_t <symbolfinish i[0]> => ('block',) + i + (t, alt_t)
     | <start> '^' <block_inner>:i
-      <template>:t <symbolfinish i[0]> => ('invertedblock',) + i + (t,)
+      <template>:t <alttemplate>:alt_t <symbolfinish i[0]> => ('invertedblock',) + i + (t,alt_t)
 alttemplate ::= (<start> <alt_inner> <template>)?:alt_t => alt_t or []
 """
 
@@ -128,34 +138,40 @@ rule ::= <literal>
     | <partial>
 block ::= [ "block" <anything>:symbol [<arg>*:arguments] [<compile>:t] [<compile>?:alt_t] ] => builder.add_block(symbol, arguments, t, alt_t)
 comment ::= [ "comment" ]
-literal ::= [ "literal" :value ] => builder.add_literal(value)
+literal ::= [ ( "literal" | "newline" | "whitespace" ) :value ] => builder.add_literal(value)
 expand ::= [ "expand" <path>:value [<arg>*:arguments]] => builder.add_expand(value, arguments)
 escapedexpand ::= [ "escapedexpand" <path>:value [<arg>*:arguments]] => builder.add_escaped_expand(value, arguments)
-invertedblock ::= [ "invertedblock" <anything>:symbol [<arg>*:arguments] [<compile>:t] ] => builder.add_invertedblock(symbol, arguments, t)
-partial ::= ["partial" <anything>:symbol [<arg>*:arguments]] => builder.add_partial(symbol, arguments)
+invertedblock ::= [ "invertedblock" <anything>:symbol [<arg>*:arguments] [<compile>:t] [<compile>?:alt_t] ] => builder.add_invertedblock(symbol, arguments, t, alt_t)
+partial ::= ["partial" <complexarg>:symbol [<arg>*:arguments]] => builder.add_partial(symbol, arguments)
 path ::= [ "path" [<pathseg>:segment]] => ("simple", segment)
- | [ "path" [<pathseg>+:segments] ] => ("complex", u'resolve(context, "'  + u'","'.join(segments) + u'")' )
-simplearg ::= [ "path" [<pathseg>+:segments] ] => u'resolve(context, "'  + u'","'.join(segments) + u'")'
+ | [ "path" [<pathseg>+:segments] ] => ("complex", u"resolve(context, '"  + u"', '".join(segments) + u"')" )
+complexarg ::= [ "path" [<pathseg>+:segments] ] => u"resolve(context, '"  + u"', '".join(segments) + u"')"
+    | [ "subexpr" ["path" <pathseg>:name] [<arg>*:arguments] ] => u'resolve_subexpr(helpers, "' + name + '", context' + (u', ' + u', '.join(arguments) if arguments else u'') + u')'
     | [ "literalparam" <anything>:value ] => {str_class}(value)
-arg ::= [ "kwparam" <anything>:symbol <simplearg>:a ] => {str_class}(symbol) + '=' + a
-    | <simplearg>
+arg ::= [ "kwparam" <anything>:symbol <complexarg>:a ] => {str_class}(symbol) + '=' + a
+    | <complexarg>
 pathseg ::= "/" => ''
     | "." => ''
     | "" => ''
     | "this" => ''
 pathseg ::= <anything>:symbol => u''.join(symbol)
-"""
+"""  # noqa: E501
 compile_grammar = compile_grammar.format(str_class=str_class.__name__)
 
 
+class PybarsError(Exception):
+
+    pass
+
+
 class strlist(list):
+
     """A quasi-list to let the template code avoid special casing."""
 
-    # Added for Python 3
-    def __str__(self):
+    def __str__(self):  # Python 3
         return ''.join(self)
 
-    def __unicode__(self):
+    def __unicode__(self):  # Python 2
         return u''.join(self)
 
     def grow(self, thing):
@@ -166,7 +182,7 @@ class strlist(list):
         # This will only ever match in Python 2 since str_class is str in
         # Python 3.
         elif type(thing) == str:
-            self.append(unicode(thing))
+            self.append(unicode(thing))  # noqa: F821 undefined name 'unicode'
 
         else:
             # Recursively expand to a flat list; may deserve a C accelerator at
@@ -183,28 +199,33 @@ _map = {
     '<': '&lt;',
     '>': '&gt;',
     }
+
+
 def substitute(match, _map=_map):
     return _map[match.group(0)]
 
 
 _escape_re = re.compile(r"&|\"|'|`|<|>")
+
+
 def escape(something, _escape_re=_escape_re, substitute=substitute):
     return _escape_re.sub(substitute, something)
 
 
 def pick(context, name, default=None):
-    if isinstance(name, str) and hasattr(context, name):
-        return getattr(context, name)
-    if hasattr(context, 'get'):
-        return context.get(name)
     try:
         return context[name]
-    except (KeyError, TypeError):
+    except (KeyError, TypeError, AttributeError):
+        if isinstance(name, str) and hasattr(context, name):
+            return getattr(context, name)
+        if hasattr(context, 'get'):
+            return context.get(name)
         return default
 
 
 sentinel = object()
 
+
 class Scope:
 
     def __init__(self, context, parent, root, overrides=None, index=None, key=None, first=None, last=None):
@@ -247,7 +268,7 @@ class Scope:
 
     # Only called in Python 2
     def __unicode__(self):
-        return unicode(self.context)
+        return unicode(self.context)  # noqa: F821 undefined name 'unicode'
 
 
 def resolve(context, *segments):
@@ -281,6 +302,44 @@ def resolve(context, *segments):
     return context
 
 
+def resolve_subexpr(helpers, name, context, *args, **kwargs):
+    if name not in helpers:
+        raise PybarsError(u"Could not find property %s" % (name,))
+    return helpers[name](context, *args, **kwargs)
+
+
+def prepare(value, should_escape):
+    """
+    Prepares a value to be added to the result
+
+    :param value:
+        The value to add to the result
+
+    :param should_escape:
+        If the string should be HTML-escaped
+
+    :return:
+        A unicode string or strlist
+    """
+
+    if value is None:
+        return u''
+    type_ = type(value)
+    if type_ is not strlist:
+        if type_ is not str_class:
+            if type_ is bool:
+                value = u'true' if value else u'false'
+            else:
+                value = str_class(value)
+        if should_escape:
+            value = escape(value)
+    return value
+
+
+def ensure_scope(context, root):
+    return context if isinstance(context, Scope) else Scope(context, context, root)
+
+
 def _each(this, options, context):
     result = strlist()
 
@@ -292,7 +351,7 @@ def _each(this, options, context):
         if last_index < 0:
             raise IndexError()
 
-    except (TypeError, IndexError) as e:
+    except (TypeError, IndexError):
         return options['inverse'](this)
 
     # We use the presence of a keys method to determine if the
@@ -312,15 +371,19 @@ def _each(this, options, context):
             value = context[value]
 
         scope = Scope(value, this, options['root'], **kwargs)
-        result.grow(options['fn'](scope))
 
+        # Necessary because of cases such as {{^each things}}test{{/each}}.
+        try:
+            result.grow(options['fn'](scope))
+        except TypeError:
+            pass
         index += 1
 
     return result
 
 
 def _if(this, options, context):
-    if isinstance(context, collections.Callable):
+    if hasattr(context, '__call__'):
         context = context(this)
     if context:
         return options['fn'](this)
@@ -345,8 +408,7 @@ def _lookup(this, context, key):
 
 
 def _blockHelperMissing(this, options, context):
-    # print this, context
-    if isinstance(context, collections.Callable):
+    if hasattr(context, '__call__'):
         context = context(this)
     if context != u"" and not context:
         return options['inverse'](this)
@@ -362,7 +424,7 @@ def _blockHelperMissing(this, options, context):
 def _helperMissing(scope, name, *args):
     if not args:
         return None
-    raise Exception(u"Could not find property %s" % (name,))
+    raise PybarsError(u"Could not find property %s" % (name,))
 
 
 def _with(this, options, context):
@@ -384,58 +446,118 @@ _pybars_ = {
 }
 
 
+class FunctionContainer:
+
+    """
+    Used as a container for functions by the CodeBuidler
+    """
+
+    def __init__(self, name, code):
+        self.name = name
+        self.code = code
+
+    @property
+    def full_code(self):
+        headers = (
+            u'import pybars\n'
+            u'\n'
+            u'if pybars.__version__ != %s:\n'
+            u'    raise pybars.PybarsError("This template was precompiled with pybars3 version %s, running version %%s" %% pybars.__version__)\n'
+            u'\n'
+            u'from pybars import strlist, Scope, PybarsError\n'
+            u'from pybars._compiler import _pybars_, escape, resolve, resolve_subexpr, prepare, ensure_scope\n'
+            u'\n'
+            u'from functools import partial\n'
+            u'\n'
+            u'\n'
+        ) % (repr(pybars.__version__), pybars.__version__)
+
+        return headers + self.code
+
+
 class CodeBuilder:
+
     """Builds code for a template."""
 
     def __init__(self):
+        self._reset()
+
+    def _reset(self):
         self.stack = []
+        self.var_counter = 1
+        self.render_counter = 0
 
     def start(self):
-        self.stack.append((strlist(), {}))
-        self._result, self._locals = self.stack[-1]
+        function_name = 'render' if self.render_counter == 0 else 'block_%s' % self.render_counter
+        self.render_counter += 1
+
+        self.stack.append((strlist(), {}, function_name))
+        self._result, self._locals, _ = self.stack[-1]
         # Context may be a user hash or a Scope (which injects '@_parent' to
         # implement .. lookups). The JS implementation uses a vector of scopes
         # and then interprets a linear walk-up, which is why there is a
         # disabled test showing arbitrary complex path manipulation: the scope
         # approach used here will probably DTRT but may be slower: reevaluate
         # when profiling.
-        self._result.grow(u"def render(context, helpers=None, partials=None, root=None):\n")
+        if len(self.stack) == 1:
+            self._result.grow([
+                u"def render(context, helpers=None, partials=None, root=None):\n"
+                u"    _helpers = dict(_pybars_['helpers'])\n"
+                u"    if helpers is not None:\n"
+                u"        _helpers.update(helpers)\n"
+                u"    helpers = _helpers\n"
+                u"    if partials is None:\n"
+                u"        partials = {}\n"
+                u"    called = root is None\n"
+                u"    if called:\n"
+                u"        root = context\n"
+            ])
+        else:
+            self._result.grow(u"def %s(context, helpers, partials, root):\n" % function_name)
         self._result.grow(u"    result = strlist()\n")
-        self._result.grow(u"    _helpers = dict(pybars['helpers'])\n")
-        self._result.grow(u"    if helpers is not None: _helpers.update(helpers)\n")
-        self._result.grow(u"    helpers = _helpers\n")
-        self._result.grow(u"    if partials is None: partials = {}\n")
-        self._result.grow(u"    if root is None: root = context\n")
-        # Expose used functions and helpers to the template.
-        self._locals['strlist'] = strlist
-        self._locals['escape'] = escape
-        self._locals['Scope'] = Scope
-        self._locals['partial'] = partial
-        self._locals['pybars'] = _pybars_
-        self._locals['resolve'] = resolve
+        self._result.grow(u"    context = ensure_scope(context, root)\n")
 
     def finish(self):
+        lines, ns, function_name = self.stack.pop(-1)
+
+        # Ensure the result is a string and not a strlist
+        if len(self.stack) == 0:
+            self._result.grow(u"    if called:\n")
+            self._result.grow(u"        result = %s(result)\n" % str_class.__name__)
         self._result.grow(u"    return result\n")
-        lines, ns = self.stack.pop(-1)
+
         source = str_class(u"".join(lines))
+
         self._result = self.stack and self.stack[-1][0]
         self._locals = self.stack and self.stack[-1][1]
-        fn = moduleFromSource(source, 'render', globalsDict=ns, registerModule=True)
-        # print source
-        return fn
 
-    def allocate_value(self, value):
-        name = 'constant_%d' % len(self._locals)
-        self._locals[name] = value
-        return name
+        code = ''
+        for key in ns:
+            if isinstance(ns[key], FunctionContainer):
+                code += ns[key].code + '\n'
+            else:
+                code += '%s = %s\n' % (key, repr(ns[key]))
+        code += source
+
+        result = FunctionContainer(function_name, code)
+        if debug and len(self.stack) == 0:
+            print('Compiled Python')
+            print('---------------')
+            print(result.full_code)
+
+        return result
 
     def _wrap_nested(self, name):
         return u"partial(%s, helpers=helpers, partials=partials, root=root)" % name
 
     def add_block(self, symbol, arguments, nested, alt_nested):
-        name = self.allocate_value(nested)
+        name = nested.name
+        self._locals[name] = nested
+
         if alt_nested:
-            alt_name = self.allocate_value(alt_nested)
+            alt_name = alt_nested.name
+            self._locals[alt_name] = alt_nested
+
         call = self.arguments_to_call(arguments)
         self._result.grow([
             u"    options = {'fn': %s}\n" % self._wrap_nested(name),
@@ -456,20 +578,16 @@ class CodeBuilder:
         self._result.grow([
             u"    value = helper = helpers.get('%s')\n" % symbol,
             u"    if value is None:\n"
-            u"        value = context.get('%s')\n" % symbol,
-            u"    if helper and callable(helper):\n"
-            u"        this = Scope(context, context, root)\n"
-            u"        value = value(this, options, %s\n" % call,
+            u"        value = resolve(context, '%s')\n" % symbol,
+            u"    if helper and hasattr(helper, '__call__'):\n"
+            u"        value = helper(context, options%s\n" % call,
             u"    else:\n"
-            u"        helper = helpers['blockHelperMissing']\n"
-            u"        value = helper(context, options, value)\n"
-            u"    if value is None: value = ''\n"
-            u"    result.grow(value)\n"
+            u"        value = helpers['blockHelperMissing'](context, options, value)\n"
+            u"    result.grow(value or '')\n"
             ])
 
     def add_literal(self, value):
-        name = self.allocate_value(value)
-        self._result.grow(u"    result.append(%s)\n" % name)
+        self._result.grow(u"    result.append(%s)\n" % repr(value))
 
     def _lookup_arg(self, arg):
         if not arg:
@@ -478,10 +596,13 @@ class CodeBuilder:
 
     def arguments_to_call(self, arguments):
         params = list(map(self._lookup_arg, arguments))
-        return u", ".join(params) + ")"
+        output = u', '.join(params) + u')'
+        if len(params) > 0:
+            output = u', ' + output
+        return output
 
     def find_lookup(self, path, path_type, call):
-        if path and path_type == "simple":  # simple names can reference helpers.
+        if path_type == "simple":  # simple names can reference helpers.
             # TODO: compile this whole expression in the grammar; for now,
             # fugly but only a compile time overhead.
             # XXX: just rm.
@@ -491,36 +612,26 @@ class CodeBuilder:
                 u"    if value is None:\n"
                 u"        value = resolve(context, '%s')\n" % path,
                 ])
-        elif path_type == "simple":
-            realname = None
-            self._result.grow([
-                u"    value = resolve(context, '%s')\n" % path,
-                ])
         else:
             realname = None
             self._result.grow(u"    value = %s\n" % path)
         self._result.grow([
-            u"    if callable(value):\n"
-            u"        this = Scope(context, context, root)\n"
-            u"        value = value(this, %s\n" % call,
+            u"    if hasattr(value, '__call__'):\n"
+            u"        value = value(context%s\n" % call,
             ])
         if realname:
             self._result.grow(
                 u"    elif value is None:\n"
-                u"        this = Scope(context, context, root)\n"
-                u"        value = helpers.get('helperMissing')(this, '%s', %s\n"
+                u"        value = helpers['helperMissing'](context, '%s'%s\n"
                     % (realname, call)
                 )
-        self._result.grow(u"    if value is None: value = ''\n")
 
     def add_escaped_expand(self, path_type_path, arguments):
         (path_type, path) = path_type_path
         call = self.arguments_to_call(arguments)
         self.find_lookup(path, path_type, call)
         self._result.grow([
-            u"    if type(value) is not strlist:\n",
-            u"        value = escape(%s(value))\n" % str_class.__name__,
-            u"    result.grow(value)\n"
+            u"    result.grow(prepare(value, True))\n"
             ])
 
     def add_expand(self, path_type_path, arguments):
@@ -528,22 +639,49 @@ class CodeBuilder:
         call = self.arguments_to_call(arguments)
         self.find_lookup(path, path_type, call)
         self._result.grow([
-            u"    if type(value) is not strlist:\n",
-            u"        value = %s(value)\n" % str_class.__name__,
-            u"    result.grow(value)\n"
+            u"    result.grow(prepare(value, False))\n"
             ])
 
     def _debug(self):
         self._result.grow(u"    import pdb;pdb.set_trace()\n")
 
-    def add_invertedblock(self, symbol, arguments, nested):
+    def add_invertedblock(self, symbol, arguments, nested, alt_nested):
         # This may need to be a blockHelperMissing clal as well.
-        name = self.allocate_value(nested)
+
+        name = nested.name
+        self._locals[name] = nested
+
+        if alt_nested:
+            alt_name = alt_nested.name
+            self._locals[alt_name] = alt_nested
+
+        call = self.arguments_to_call(arguments)
         self._result.grow([
-            u"    value = context.get('%s')\n" % symbol,
-            u"    if not value:\n"
-            u"    "])
-        self._invoke_template(name, "context")
+            u"    options = {'inverse': %s}\n" % self._wrap_nested(name),
+            u"    options['helpers'] = helpers\n"
+            u"    options['partials'] = partials\n"
+            u"    options['root'] = root\n"
+            ])
+        if alt_nested:
+            self._result.grow([
+                u"    options['fn'] = ",
+                self._wrap_nested(alt_name),
+                u"\n"
+                ])
+        else:
+            self._result.grow([
+                u"    options['fn'] = lambda this: None\n"
+                ])
+        self._result.grow([
+            u"    value = helper = helpers.get('%s')\n" % symbol,
+            u"    if value is None:\n"
+            u"        value = resolve(context, '%s')\n" % symbol,
+            u"    if helper and hasattr(helper, '__call__'):\n"
+            u"        value = helper(context, options%s\n" % call,
+            u"    else:\n"
+            u"        value = helpers['blockHelperMissing'](context, options, value)\n"
+            u"    result.grow(value or '')\n"
+            ])
 
     def _invoke_template(self, fn_name, this_name):
         self._result.grow([
@@ -567,7 +705,8 @@ class CodeBuilder:
                         overrides = {}
                     overrides[kwmatch.group(1)] = kwmatch.group(2)
                 else:
-                    assert positional_args == 0, positional_args
+                    if positional_args != 0:
+                        raise PybarsError("An extra positional argument was passed to a partial")
                     positional_args += 1
                     arg = argument
 
@@ -580,12 +719,16 @@ class CodeBuilder:
         self._result.grow([u"    overrides = %s\n" % overrides_literal])
 
         self._result.grow([
-            u"    inner = partials['%s']\n" % symbol,
+            u"    partialName = %s\n" % symbol,
+            u"    if partialName not in partials:\n",
+            u"        raise PybarsError('The partial %s could not be found' % partialName)\n",
+            u"    inner = partials[partialName]\n",
             u"    scope = Scope(%s, context, root, overrides=overrides)\n" % self._lookup_arg(arg)])
         self._invoke_template("inner", "scope")
 
 
 class Compiler:
+
     """A handlebars template compiler.
 
     The compiler is not threadsafe: you need one per thread because of the
@@ -598,24 +741,200 @@ class Compiler:
 
     def __init__(self):
         self._helpers = {}
+        self.template_counter = 1
 
-    def compile(self, source):
+    def _extract_word(self, source, position):
+        """
+        Extracts the word that falls at or around a specific position
+
+        :param source:
+            The template source as a unicode string
+        :param position:
+            The position where the word falls
+
+        :return:
+            The word
+        """
+        boundry = re.search('{{|{|\s|$', source[:position][::-1])
+        start_offset = boundry.end() if boundry.group(0).startswith('{') else boundry.start()
+
+        boundry = re.search('}}|}|\s|$', source[position:])
+        end_offset = boundry.end() if boundry.group(0).startswith('}') else boundry.start()
+
+        return source[position - start_offset:position + end_offset]
+
+    def _generate_code(self, source):
+        """
+        Common compilation code shared between precompile() and compile()
+
+        :param source:
+            The template source as a unicode string
+
+        :return:
+            A tuple of (function, source_code)
+        """
+
+        if not isinstance(source, str_class):
+            raise PybarsError("Template source must be a unicode string")
+
+        tree, (position, _) = self._handlebars(source).apply('template')
+
+        self.clean_whitespace(tree)
+
+        if debug:
+            print('\nAST')
+            print('---')
+            print(tree)
+            print('')
+
+        if position < len(source):
+            line_num = source.count('\n') + 1
+            beginning_of_line = source.rfind('\n', 0, position)
+            if beginning_of_line == -1:
+                char_num = position
+            else:
+                char_num = position - beginning_of_line
+            word = self._extract_word(source, position)
+            raise PybarsError("Error at character %s of line %s near %s" % (char_num, line_num, word))
+
+        # Ensure the builder is in a clean state - kinda gross
+        self._compiler.globals['builder']._reset()
+
+        output = self._compiler(tree).apply('compile')[0]
+        return output
+
+    def precompile(self, source):
+        """
+        Generates python source code that can be saved to a file for caching
+
+        :param source:
+            The template to generate source for - should be a unicode string
+
+        :return:
+            Python code as a unicode string
+        """
+
+        return self._generate_code(source).full_code
+
+    def compile(self, source, path=None):
         """Compile source to a ready to run template.
 
-        :param source: The template to compile - should be a unicode string.
-        :return: A template ready to run.
-        """
-        assert isinstance(source, str_class)
-        tree = self._handlebars(source).apply('template')[0]
-        # print source
-        # print '-->'
-        # print "T", tree
-        code = self._compiler(tree).apply('compile')[0]
-        # print code
-        return code
+        :param source:
+            The template to compile - should be a unicode string
 
-#orig = Compiler._handlebars.rule_blockrule
-#def thunk(*args, **kwargs):
-#    import pdb;pdb.set_trace()
-#    return orig(*args, **kwargs)
-#Compiler._handlebars.rule_blockrule = thunk
+        :return:
+            A template function ready to execute
+        """
+
+        container = self._generate_code(source)
+
+        def make_module_name(name, suffix=None):
+            output = 'pybars._templates.%s' % name
+            if suffix:
+                output += '_%s' % suffix
+            return output
+
+        if not path:
+            path = '_template'
+            generate_name = True
+        else:
+            path = path.replace('\\', '/')
+            path = path.replace('/', '_')
+            mod_name = make_module_name(path)
+            if mod_name in sys.modules:
+                generate_name = True
+
+        if generate_name:
+            mod_name = make_module_name(path, self.template_counter)
+            while mod_name in sys.modules:
+                self.template_counter += 1
+                mod_name = make_module_name(path, self.template_counter)
+
+        mod = ModuleType(mod_name)
+        filename = '%s.py' % mod_name.replace('pybars.', '').replace('.', '/')
+        exec(compile(container.full_code, filename, 'exec', dont_inherit=True), mod.__dict__)
+        sys.modules[mod_name] = mod
+        linecache.getlines(filename, mod.__dict__)
+
+        return mod.__dict__[container.name]
+
+    def clean_whitespace(self, tree):
+        """
+        Cleans up whitespace around block open and close tags if they are the
+        only thing on the line
+
+        :param tree:
+            The AST - will be modified in place
+        """
+
+        pointer = 0
+        end = len(tree)
+
+        while pointer < end:
+            piece = tree[pointer]
+            if piece[0] == 'block':
+                child_tree = piece[3]
+
+                # Look at open tag, if the only other thing on the line is whitespace
+                # then delete it so we don't introduce extra newlines to the output
+                open_pre_whitespace = False
+                open_pre_content = True
+                if pointer > 1 and tree[pointer - 1][0] == 'whitespace' and (tree[pointer - 2][0] == 'newline' or tree[pointer - 2] == 'template'):
+                    open_pre_whitespace = True
+                    open_pre_content = False
+                elif pointer > 0 and (tree[pointer - 1][0] == 'newline' or tree[pointer - 1] == 'template'):
+                    open_pre_content = False
+
+                open_post_whitespace = False
+                open_post_content = True
+                child_len = len(child_tree)
+                if child_len > 2 and child_tree[1][0] == 'whitespace' and child_tree[2][0] == 'newline':
+                    open_post_whitespace = True
+                    open_post_content = False
+                elif child_len > 1 and child_tree[1][0] == 'newline':
+                    open_post_content = False
+
+                if not open_pre_content and not open_post_content:
+                    if open_pre_whitespace:
+                        tree.pop(pointer - 1)
+                        pointer -= 1
+                        end -= 1
+                    if open_post_whitespace:
+                        child_tree.pop(1)
+                    child_tree.pop(1)  # trailing newline
+
+                # Do the same thing, but for the close tag
+                close_pre_whitespace = False
+                close_pre_content = True
+                child_len = len(child_tree)
+                if child_len > 2 and child_tree[child_len - 1][0] == 'whitespace' and child_tree[child_len - 2][0] == 'newline':
+                    close_pre_whitespace = True
+                    close_pre_content = False
+                elif child_len > 1 and child_tree[child_len - 1][0] == 'newline':
+                    close_pre_content = False
+
+                close_post_whitespace = False
+                close_post_content = True
+                tree_len = len(tree)
+                if tree_len > pointer + 2 and tree[pointer + 1][0] == 'whitespace' and tree[pointer + 2][0] == 'newline':
+                    close_post_whitespace = True
+                    close_post_content = False
+                elif tree_len == pointer + 2 and tree[pointer + 1][0] == 'whitespace':
+                    close_post_whitespace = True
+                    close_post_content = False
+                elif tree_len > pointer + 1 and tree[pointer + 1][0] == 'newline':
+                    close_post_content = False
+                elif tree_len == pointer + 1:
+                    close_post_content = False
+
+                if not close_pre_content and not close_post_content:
+                    if close_pre_whitespace:
+                        child_tree.pop()
+                    child_tree.pop()  # preceeding newline
+                    if close_post_whitespace:
+                        tree.pop(pointer + 1)
+                        end -= 1
+
+                self.clean_whitespace(child_tree)
+
+            pointer += 1
